/*
 * keydb.c
 *
 * Part of the strings db toolkit.
 *
 * This is the implementation for the functions prototyped in strdb.h.
 * It's part of a miniature library for the toolkit.
 *
 * For usage of the various functions, read the comments preceding each.
 *
 * 0.3.0, March 23, 1997.  Keys.  Sorting moved to sortdb.c
 * 0.4.0, March 24, 1997.  No mods.
 *
 * Copyright (c) 1996 Amelia A. Lewis
 * Distributed under GPL.  See COPYING.
 *
 * $Id: keydb.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include "keydb.h"

/*
 * sdb_key_write
 */
int
sdb_key_write(sdb_key *k, FILE *keyf)
{
    unsigned char slen;

    if (!k || !keyf)
        return 0;
    if (fseek(keyf, 0, SEEK_END))
        return -2;

    slen = (unsigned char)strlen(k->key);
    if (!fwrite((unsigned char *)&slen, sizeof(unsigned char), 1, keyf))
        return -1;
    if (!fwrite((char *)k->key, sizeof(char), strlen(k->key), keyf))
        return -1;
    if (!fwrite((unsigned short *)&k->nindices, sizeof(unsigned short), 1, keyf))
        return -1;
    if (!fwrite((unsigned short *)k->indices, sizeof(unsigned short), k->nindices, keyf))
        return -1;
    return k->nindices;
}

/*
 * sdb_key_read
 *
 * Reads a key into the first parameter k from the current position in
 * the named file keyf, which must be an open keyfile positioned at
 * the start of a key.
 * Returns the pointer to the character string of the key (the key member
 * of the sdb_key structure).
 *
 * Storage for the key is allocated on the heap.
 */
char *
sdb_key_read(sdb_key *k, FILE *keyf)
{
    unsigned char str_length;
    
    if (!k || !keyf)
        return NULL;
    if (!fread((unsigned char *)&str_length, sizeof(unsigned char), 1, keyf))
        return NULL;
    if ((k->key = (char *)calloc(str_length + 1, sizeof(char))) == NULL)
        return NULL;
    if (!fread((char *)k->key, sizeof(char), str_length, keyf))
    {
        free(k->key);
        return NULL;
    }
    k->key[str_length] = '\0';
    if (!fread((unsigned short *)&k->nindices, sizeof(unsigned short), 1, keyf))
    {
        free(k->key);
        return NULL;
    }
    if ((k->indices = (unsigned short *)calloc(k->nindices, sizeof(unsigned short))) == NULL)
    {
        free (k->key);
        k->nindices = 0;
        return NULL;
    }
    if (!fread((unsigned short *)k->indices, sizeof(unsigned short), k->nindices, keyf))
    {
        free(k->indices);
        k->nindices = 0;
        free(k->key);
        return NULL;
    }
    return k->key;
}

/*
 * sdb_key_free
 *
 * Properly deallocates storage allocated by sdb_key_read.
 */
void
sdb_key_free(sdb_key *k)
{
    free(k->key);
    k->key = NULL;
    free(k->indices);
    k->indices = NULL;
    k->nindices = 0;
}

/*
 * sdb_key_match
 *
 * Simply compares a key, as stored in the key file and retrieved with
 * sdb_key_read, with a character string passed in kcomp.  The string
 * in kcomp will have whitespace removed, and will be lowercased; this
 * is done automatically for stored keys.
 *
 * I don't see much reason for using regexes with keys.  Maybe later.
 */
int
sdb_key_match(sdb_key *k, char *kcomp)
{
    int i;
    
    if (!k || !kcomp)
        return 0;
    for (i = 0; i < strlen(kcomp); i++)
        kcomp[i] = tolower(kcomp[i]);
    return(!strcmp(k->key, kcomp));
}
