/*
 * sortdb.c
 *
 * Part of the strings db toolkit.
 *
 * Distributed under GPL.  See COPYING.
 * sdb_llmsort derived from the public domain llmsort by Philip J. Erdelsky.
 *
 * $Id: sortdb.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include "strdb.h"
#include "sortdb.h"

/*
 * sdb_comparefield:
 *
 * A simple compare function (no, really!) that does a strcmp on a given
 * field.  Compares the records pointed to by the two pointers, finding
 * fields with the signal character s and the field delimiter f.  To
 * check the ambiguous field, f must return false for isgraph().
 * cflag is a case flag; if true, case is significant.
 * wflag is a whitespace flag.  If 0, whitespace is significant.  If
 * 1, whitespace is not significant, but control characters are.  If
 * greater than one, only alphanumerics are significant (isalnum()).
 */
int 
sdb_comparefield(sdb_recptr *p1, sdb_recptr *p2, unsigned char flags,
                 char s, char f, FILE *d)
{
    char *f1, *f2, *fend;
    int x;
    unsigned long u1, u2;
    
    if (!(d) || !(p1) || !(p2))
        return 0;
    if (!(sdb_recalloc(d, p1)) || !(sdb_recalloc(d, p2)))
    {
        if (p1->data) sdb_recfree(p1);
        if (p2->data) sdb_recfree(p2);
        return 0;
    }
    if (!isgraph(s))
        return 0;
    if (isgraph(f))
    {
        f1 = sdb_finr(p1->data, s, f);
        f2 = sdb_finr(p2->data, s, f);
    }
    else
    {
        f1 = p1->data;
        f2 = p2->data;
    }
    if (!(f1))
    {
        if (!(f2)) return 0; /* no field in either */
        else return -1; /* no f1, so less than f2, which exists */
    }
    else if (!(f2)) return 1; /* f1 exists, so greater than f2 which doesn't */
    /* now truncate */
    fend = sdb_ndinr(f1, s);
    *fend = '\0';
    fend = sdb_ndinr(f2, s);
    *fend = '\0';
    if (!(flags & sdb_SORT_CASE)) /* ignore case */
    {
        for (fend = f1 ; *fend ; fend++)
            *fend = tolower(*fend);
        for (fend = f2 ; *fend ; fend++)
            *fend = tolower(*fend);
    }
    if (flags & sdb_SORT_IWHITE) /* ignore whitespace */
    {
        for (fend = f1 ; *fend && !(isgraph(*fend)) ; fend++); /* no body */
        f1 = fend;
        for (fend = f2 ; *fend && !(isgraph(*fend)) ; fend++);
        f2 = fend;
    }
    if (flags & sdb_SORT_ALNUM)
    {
        for (fend = f1 ; *fend && !(isalnum(*fend)) ; fend++);
        f1 = fend;
        for (fend = f2 ; *fend && !(isalnum(*fend)) ; fend++);
        f2 = fend;
    }
    if (flags & sdb_SORT_NUMBER)
    {
        for (fend = f1 ; *fend && !(isdigit(*fend)) ; fend++);
        f1 = fend;
        for (fend = f2 ; *fend && !(isdigit(*fend)) ; fend++);
        f2 = fend;
        /* this one finishes here, others in the else */
        u1 = u2 = 0;
        if (*f1)
            u1 = strtoul(f1, NULL, 10);
        if (*f2)
            u2 = strtoul(f2, NULL, 10);
        x = (u1 > u2) ? 1 : ((u1 == u2) ? 0 : -1);
    }
    else
    {
        x = strcmp(f1, f2);  /* all that trouble for this little thing */
    }
    if (p1->data) sdb_recfree(p1);
    if (p2->data) sdb_recfree(p2);
    return x;
}

/*
 * sdb_llmsort:
 *
 * Based on the llmsort released into the public domain by Philip J.
 * Erdelsky.  Iterative, not recursive.  The two parameters important to
 * the function itself are startp, the start of a linked list of rec_ptr
 * pointers, and the comparison function.  Applications may use the 
 * sort with their own comparison function (they have to correctly create
 * the header, too, of course).  The final three parameters are passed
 * to the comparison function, which must accept two pointers to sdb_recptr,
 * the signal character, the field on which to sort, and the strings
 * file.
 * Added a case flag, an ignore whitespace flag, and an alphanum only flag.
 * All get passed to the comparison function.
 *
 * To sort on the ambiguous field, pass in fld = '\0'.
 *
 * On completion, returns a pointer to the first record of the sorted
 * list.
 */
sdb_recptr *
sdb_llmsort(sdb_recptr *startp, int (*compfunc)(sdb_recptr *, sdb_recptr *,
                        unsigned char, char, char, FILE *), unsigned char flg,
                        char sc, char fld, FILE *df, int verbose)
{
    struct tape_struct
    {
        sdb_recptr *start, *end;
        unsigned long count;
    } tape[4], *ctape, *otape, *tape0, *tape1;
    sdb_recptr *nextp, *chosenp;
    unsigned base, block_size, n0, n1;
    int dest;

    if (verbose)
        fprintf(stderr, "Sorting ");    
    if (!(df) || !(startp))
    {
        if (verbose)
            fprintf(stderr, "done\n");
        return NULL;
    }
    if (!(startp->next))
    {
        if (verbose)
            fprintf(stderr, "done\n");
        return startp;
    }
    
    base = tape[0].count = tape[1].count = 0;
    
    do
    {
        nextp = (sdb_recptr *)startp->next;
        startp->next = tape[base].start;
        tape[base].start = startp;
        tape[base].count++;
        startp = nextp;
        base ^= 1;
    } while (startp);
    if (!(tape[0].count)) return NULL;
    if (!(tape[1].count)) return tape[0].start;
    for (base = 0, block_size = 1; tape[base+1].count ; base ^= 2, block_size <<= 1)
    {
        if (verbose)
            fprintf(stderr, ".");
        tape0 = tape + base;
        tape1 = tape + base + 1;
        dest = base ^ 2;
        tape[dest].count = tape[dest+1].count = 0;
        for ( ; tape0->count ; dest ^= 1)
        {
            otape = tape + dest;
            n0 = n1 = block_size;
            while (1)
            {
                if (!(n0) || !(tape0->count))
                {
                    if (!(n1) || !(tape1->count)) break;
                    ctape = tape1;
                    n1--;
                }
                else if (!(n1) || !(tape1->count))
                {
                    ctape = tape0;
                    n0--;
                }
                else if ((*compfunc)(tape0->start, tape1->start, 
                                     flg, sc, fld, df) > 0)
                {
                    ctape = tape1;
                    n1--;
                }
                else
                {
                    ctape = tape0;
                    n0--;
                }
                ctape->count--;
                chosenp = ctape->start;
                ctape->start = (sdb_recptr *)chosenp->next;
                if (!(otape->count)) otape->start = chosenp;
                else otape->end->next = chosenp;
                otape->end = chosenp;
                otape->count++;
            }
        }
    }
    tape[base].end->next = NULL;
    
    if (flg & sdb_SORT_REVERSE)
    {
        if (verbose)
            fprintf(stderr, " reversing order");
        tape[base].start = sdb_reverse_order(tape[base].start);
    }
    
    if (verbose)
        fprintf(stderr, " done\n");
    return tape[base].start;
}

/*
 * sdb_reverse_order
 *	Reverses the order of a list of sdb_recptrs.
 */
sdb_recptr *
sdb_reverse_order(sdb_recptr *startp)
{
    sdb_recptr *tempp, *prevnext;

    if (startp && startp->next)
    {
        /* set up the start of the list as the new end of list */
        prevnext = startp;
        tempp = (sdb_recptr *)startp->next;
        startp = tempp;
        prevnext->next = NULL;
        while (tempp != prevnext)
        {
            if (tempp->next)		/* not the end of the list */
                startp = (sdb_recptr *)tempp->next;
            tempp->next = prevnext;
            prevnext = tempp;
            tempp = startp;
        }
    }
    else /* there's only one pointer in the list, or it's empty */
    {
        tempp = startp;
    }
    return tempp;
}
