/*
 * strdb.c
 *
 * Part of the strings db toolkit.
 *
 * This is the implementation for the functions prototyped in strdb.h.
 * It's used as a miniature library for the toolkit.
 *
 * For usage of the various functions, read the comments preceding each.
 * Or check to see if the manpage is finished.
 *
 * Copyright (c) 1996 Amy A. Lewis
 * Distributed under GPL.  See COPYING.
 *
 * $Id: strdb.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#ifdef USE_REGEX
#include <regex.h>
#endif /* USE_REGEX */
#include "strdb.h"
#include "errdb.h"
#include "version.h"

const char sdb_VMAJOR = VERSION_MAJOR;
const char sdb_VMINOR = VERSION_MINOR;
const char sdb_VPATCH = VERSION_PATCH;
const char sdb_DEFSC = '%';
const char sdb_MAGIC[4] = "aal";
const unsigned short MAXLINELEN = 1024;		/* can we get *rid* of this? */
const sdb_record_t sdb_RECORDMAX = USHRT_MAX;

/* Have to have sorting constants here, for decoding */
const unsigned char sdb_SORT_CASE	= (1 << 0);
const unsigned char sdb_SORT_REVERSE	= (1 << 1);
const unsigned char sdb_SORT_IWHITE	= (1 << 2);
const unsigned char sdb_SORT_ALNUM	= (1 << 3);
const unsigned char sdb_SORT_NUMBER	= (1 << 4);


/*
 * sdb_pointer
 *	Returns the pointer indicated by the index number which.
 * Returns a newly allocated sdb_recptr structure, which must be freed by
 * the caller.  The structure has only its offset and size members initialized.
 * On error, it returns NULL.
 */
sdb_recptr *
sdb_pointer(FILE *idxf, unsigned long which)
{
    sdb_recptr *retval;
    int success = 1;
    
    if (!(idxf))
        return NULL;

    if ((retval = (sdb_recptr *)malloc(sizeof(sdb_recptr))) != NULL)
    {
        retval->data = NULL;
        retval->next = NULL;
        fseek(idxf, sizeof(sdb_header) + ((which - 1) * (sizeof(off_t) + sizeof(unsigned short))), SEEK_SET);
        if (!fread((off_t *)&retval->data_os, sizeof(off_t), 1, idxf))
            success = 0;
        else
            if (!fread((unsigned short *)&retval->data_size, sizeof(unsigned short), 1, idxf))
                success = 0;
    }
    if (!success)
    {
        free(retval);
        retval = NULL;
    }
    return retval;
}

/*
 * sdb_nextptr
 *	Returns a newly allocated sdb_recptr structure, which the caller
 * must free.  Wraps at eof.  Returns NULL on error.  Only offset and
 * and size are initialized.
 */
sdb_recptr *
sdb_nextptr(FILE *idxf)
{
    sdb_recptr *retval;
    int success = 1;
    
    if (!(idxf))
        return NULL;

    if ((retval = (sdb_recptr *)malloc(sizeof(sdb_recptr))) != NULL)
    {
        retval->next = NULL;
        retval->data = NULL;
        if (!fread((off_t *)&retval->data_os, sizeof(off_t), 1, idxf))
            success = 0;
        else
            if (!fread((unsigned short *)&retval->data_size, sizeof(unsigned short), 1, idxf))
                success = 0;
        if (feof(idxf))
        {
            fseek(idxf, sizeof(sdb_header), SEEK_SET);
            if (fread((off_t *)&retval->data_os, sizeof(off_t), 1, idxf))
            {
                success = 1;	/* else there are *no* pointers */
                fread((unsigned short *)&retval->data_size, sizeof(unsigned short), 1, idxf);
            }
        }
    }
    if (!success)
    {
        free(retval);
        retval = NULL;
    }
    return retval;
}

/*
 * sdb_prevptr
 *	Returns the address of the previous pointer, wrapping at start of
 * file.
 */
sdb_recptr *
sdb_prevptr(FILE *idxf)
{
    sdb_recptr *retval;
    int success = 1;
    
    if (!(idxf))
        return NULL;

    fseek(idxf, -2 * (sizeof(off_t) + sizeof(unsigned short)), SEEK_CUR);
    if (ftell(idxf) < sizeof(sdb_header))
        fseek(idxf, -1 * (sizeof(off_t) + sizeof(unsigned short)), SEEK_END);
    if (ftell(idxf) < sizeof(sdb_header))
        return NULL;			/* no records at all */

    if ((retval = (sdb_recptr *)malloc(sizeof(sdb_recptr))) != NULL)
    {
        retval->data = NULL;
        retval->next = NULL;
        if (!fread((off_t *)&retval->data_os, sizeof(off_t), 1, idxf))
           success = 0;
        else
            if (!fread((unsigned short *)&retval->data_size, sizeof(unsigned short), 1, idxf))
                success = 0;
    }
    if (!success)
    {
        free(retval);
        retval = NULL;
    }
    return retval;
}

/*
 * sdb_delim
 * Given a null-terminated array of characters, and the current signal
 * character, returns 1 if the line is a record terminator, and the field
 * id character if the line is a field start character.  If neither is
 * true, returns 0.
 */
char 
sdb_delim(char *linep, char sc)
{
    if (!(linep))
        return 0;
    if (!(isgraph(sc)))
        return 0; /* only printable characters allowed */
    if ( (linep[0] != sc) /* doesn't start with sig char, commonest case */
          || (strlen(linep) > 3) /* too long, second most common */
          || (linep[strlen(linep) -1] != '\n') ) /* doesn't end with newline ?? */
        return 0;

    /* only lines 2 and 3 chars long, with only one char undefined, remain */
    if (strlen(linep) == 2) return 1;	/* ambiguous field, sig char only */
    if (linep[1] == sc) return 1;	/* ambiguous field, two sig chars */
    if (isgraph(linep[1]))
        return linep[1];		/* named field */
    if (linep[1] == '\b')
        return linep[1];		/* deleted record */

    return 0;				/* unprintable character, not a field */
}

/*
 * sdb_ndinr
 * Finds the next delimiter in a record.  Returns NULL only if the record
 * pointer is null; otherwise returns a pointer to the next field, or a
 * pointer to the terminating null in the record.
 *
 * This is recursive, but shouldn't be a problem unless someone uses an
 * awful lot of faked field delimiters.
 */
char *
sdb_ndinr(char *rec, char schar)
{
    char *fnext, fdelim[3];
    
    if (!(rec))
        return NULL;

    sprintf(fdelim, "\n%c", schar);
    fnext = strstr(rec,fdelim);
    if (!(fnext))
        return(rec + strlen(rec));		/* null at eor */
    if ((fnext + 3) >= (rec + strlen(rec)))
        return fnext;				/* it's gonna be empty */
    if ((fnext[3] == '\n') && isgraph(fnext[2]))
        return fnext;				/* normal field */
    else 
        return(sdb_ndinr(fnext + 2, schar));	/* it wasn't really a delim */
}

/*
 * sdb_finr:
 * Field in record.  Returns a pointer to the specified field in the record
 * passed in, or NULL if no field is found.
 *
 * Usage note: to find the *next* field in a record, pass the pointer to
 * the linefeed that ends the field just found as the first parameter.
 */
char *
sdb_finr(char *rec, char sc, char f)
{
    char delim[4];
    
    if (!(rec))
        return NULL;
    sprintf(delim, "%c%c\n", sc, f);
    return(strstr(rec, delim));
}

/*
 * sdb_findindex
 * A function to find the appropriate index file from a perhaps-incomplete
 * filename.  Returns NULL on failure; returns the type of error in the
 * pointer to int passed in.  Cleanup of filename and error pointer are
 * the responsibility of the caller.  A value returned in the error code
 * means that the file exists, but is not an index file.
 *
 * Revision: the char * passed in should be a buf[MAXPATHLEN].  It will
 * receive the real name of the file.
 *
 * The final parameter, which should be zero or 1, indicates whether the
 * file should be opened for writing as well as reading.
 *
 * ecode will contain an sdb_unicorn error code.  If gibber is on, an
 * outdated file will be used, and the error code returned will be 0 (no
 * error).
 */
FILE *
sdb_findindex(char *fname, int *ecode, int gibber, int rplus)
{
    char *tstr = NULL, exten[8], i, moder[3];
    FILE *retf = NULL;
    
    if (!fname || !ecode)
        return NULL;
    *ecode = 0;
    if (rplus)
        strcpy(moder, "r+");
    else
        strcpy(moder, "r");

    /* check for period in filename.  If yes, try that as index. */
    if (strchr(fname, '.'))
    {
        /* yes.  return fname */
        retf = fopen(fname, moder);
        if (retf)
        {
            *ecode = sdb_unicorn(retf, 0);
            if (gibber && (*ecode == sdb_EIDXDATE)) 
                *ecode = 0;
            else if (*ecode)
            {
                fclose(retf);
                retf = NULL;
            }
        }
    }
    /* no.  Try with extension */
    if (!retf)
    {
        strcpy(exten, ".index");
        tstr = (char *)malloc(strlen(fname)+strlen(exten));
        if (!tstr)
        {
            *ecode = sdb_EMEMFAULT;
            return NULL;		/* memory allocation error */
        }
        strcpy(tstr, fname);
        strcat(tstr, exten);
        retf = fopen(tstr, moder);
        if (retf)
        {
            *ecode = sdb_unicorn(retf, 0);
            if (*ecode == sdb_EIDXDATE)	/* right file, wrong date */
            {
                if (gibber)
                {
                    strcpy(fname, tstr);
                    *ecode = 0;
                }
                else
                    return NULL;
            }
            else if (*ecode)
            {
                fclose(retf);
                retf = NULL;
            }
        }
        /* sigh.  Try all the *other* extensions, in order */
        if (!retf)
            for (i = 0x21 ; i < 0x7F ; i++) /* range of isgraphic */
            {
                strcpy(tstr, fname);
                sprintf(exten, ".%cdx", i);
                strcat(tstr, exten);
                retf = fopen(tstr, moder);
                if (retf)
                {
                    *ecode = sdb_unicorn(retf, 0);
                    if (gibber && (*ecode == sdb_EIDXDATE)) /* use outdated file */
                    {
                        strcpy(fname, tstr);
                        *ecode = 0;
                        break;
                    }
                    else if (*ecode)
                    {
                        fclose(retf);
                        retf = NULL;
                    }
                }
            }
        free(tstr);
        /* sigh.  Try *without* extension, just like above. */
        if (!retf)
        {
            retf = fopen(fname, moder);
            if (retf)
            {
                *ecode = sdb_unicorn(retf, 0);
                if (gibber && (*ecode == sdb_EIDXDATE)) 
                    *ecode = 0;
                else
                {
                    fclose(retf);
                    retf = NULL;
                }
            }
        }
    }
    return retf;
}

/*
 * sdb_recalloc:
 * One of the commonest functions.  This takes the (open) data file, 
 * an sdb_recptr which has the offset and size fields already filled
 * in, and allocates memory for the data.
 * It returns the size of the buffer on success, or -1 on error.  The
 * buffer may be zero length.
 */
int
sdb_recalloc(FILE *df, sdb_recptr *sr)
{
    if (!(df) || !(sr))
        return -1;
    if (fseek(df, sr->data_os, SEEK_SET) == -1)
        return -1;

    if ((sr->data = (char *)calloc(sr->data_size + 1, sizeof(char))) != NULL)
    {
        sr->data[sr->data_size] = '\0';
        if (sr->data_size != fread((char *)sr->data, sizeof(char), sr->data_size, df))
        {
            free(sr->data);
            sr->data = NULL;
            return -1;
        }
    }
    
    return(sr->data ? sr->data_size : -1);
}

/*
 * sdb_stripterm
 *	Remove the terminator from the end of an allocated record, updating
 * the size field appropriately.  Returns number of characters removed
 * (should be 3 or 2, but can be 0 at EOF).
 */
int
sdb_stripterm(sdb_recptr *rec, char sigc)
{
    if (!rec || !rec->data)
        return 0;
    
    if ((rec->data[rec->data_size - 1] == '\n') && 
        ((rec->data[rec->data_size - 2] == sigc) ||
        (rec->data[rec->data_size - 2] == '\b')))
    {
        if ((rec->data[rec->data_size - 3] == '\n') &&
            (rec->data[rec->data_size - 2] == sigc))
        {
            rec->data[rec->data_size - 2] = '\0';
            rec->data_size -= 2;
            return 2;
        }
        if ((rec->data[rec->data_size - 3] == sigc) &&
            (rec->data[rec->data_size - 4] == '\n'))
        {
            rec->data[rec->data_size - 3] = '\0';
            rec->data_size -= 3;
            return 3;
        }
    }
    if ((rec->data[rec->data_size - 1] == sigc) ||
        (rec->data[rec->data_size - 1] == '\b'))/* degenerate case */
    {
        if ((rec->data[rec->data_size - 2] == '\n') &&
            (rec->data[rec->data_size - 1] == sigc))
        {
            rec->data[rec->data_size - 1] = '\0';
            rec->data_size--;
            return 1;
        }
        if ((rec->data[rec->data_size - 2] == sigc) &&
            (rec->data[rec->data_size - 3] == '\n'))
        {
            rec->data[rec->data_size - 2] = '\0';
            rec->data_size -= 2;
            return 2;
        }
    }
    return 0;
}

/*
 * sdb_recfree
 *	Frees the memory holding the data of the record.
 */
void
sdb_recfree(sdb_recptr *liberation)
{
    free(liberation->data);
    liberation->data = NULL;
}

/*
 * sdb_grep
 *	Yes, it's sort of messy, isn't it?
 * Okay.  Takes a file pointer to the data file, the pointer to the index
 * file (we could get the data file from the index file, but it saves time
 * to have it open already), a null-delimited string to match (which may
 * contain regex characters), the field to search in, an array big enough
 * to hold as many records as we're going to return, the number to indicate
 * how many that is, with 0 indicating that we want *all* records, 
 * a flag to indicate whether this is a case-sensitive match, a flag to
 * indicate whether we return the index numbers of the pointers or the
 * contents of the indexed pointer, and finally a flag to indicate whether
 * this is to be treated as a regular expression (valid only if compiled
 * with some sort of POSIX compatible regex library, which includes the
 * GNU c library, but might not include other folks' c libraries).
 *
 * It returns a count of the number of values returned.  Those values
 * are returned in osarray, which damned well better be big enough to
 * hold them.  They can be either the index numbers or offsets (pointers).
 * The index numbers are one-based, to use with the pointer functions.
 *
 * The bloody function includes variables to hold the record, to point at
 * the searched field within the record, and to point at the searched field's
 * end.  There's a header.  There's a long to be the current index number,
 * and another to be the counter of matched records.  And of course there's
 * a temporary thing to hold the current offset.  There's a thingy to hold
 * the compiled regular expression, if there is one, and a pointer at the
 * function to use to movebackward or forward.
 */
unsigned long 
sdb_grep(FILE *df, FILE *idxf, char *match, char fld, off_t *osarray, 
         unsigned long maxreturn, int caseflag, int indexflag, int rxflag,
         int all_flag, int last_flag)
{
    char *rstart, *fstart, *fend;
    unsigned long l, counter = 0;
    sdb_header sdbh;
    sdb_recptr *curptr = NULL;
#ifdef USE_REGEX
    regex_t *compreg = NULL;
#endif
    sdb_recptr *(*ptr_move)(FILE *);
    int finished = 0;

    if (!(idxf) || !(df) || !(match)) return 0;
    if (!last_flag)
        ptr_move = sdb_prevptr;
    else
        ptr_move = sdb_nextptr;

#ifdef USE_REGEX
    if (rxflag)
    {
        compreg = (regex_t *)malloc(sizeof(regex_t));
        if (!compreg)
            return 0;
        if (regcomp(compreg, match, REG_EXTENDED | (caseflag ? 0 : REG_ICASE | REG_NOSUB)))
        {
            regfree(compreg);
            free(compreg);
            return 0;
        }
    }
#endif

    rewind(idxf);
    rewind(df);
    fread((sdb_header *)&sdbh, sizeof(sdb_header), 1, idxf);

    for (l = 1; (l <= sdbh.sdb_total) && (counter < (maxreturn ? maxreturn : sdbh.sdb_total)); l++)
    {
        curptr = ptr_move(idxf);	/* we'll catch NULL in next line */
        if (sdb_recalloc(df, curptr) > 0)
        {
            rstart = curptr->data;
            while (!finished)
            {
                if (fld)
                {
                    fstart = sdb_finr(rstart, sdbh.sdb_sigch, fld);
                    if (fstart)
                        fstart += 3;
                    rstart = fstart;
                }
                else 
                {
                    fstart = rstart;
                    finished++;
                }
                if (fstart)
                {
                    if (!all_flag)
                    {
                        fend = sdb_ndinr(fstart, sdbh.sdb_sigch);
                        *fend = '\0';
                    }
#ifdef USE_REGEX
                    if (rxflag)
                    {
                        if (!regexec(compreg, fstart, 0, NULL, 0))
                        {
                            if (indexflag != 1)
                                osarray[counter] = (last_flag ? (off_t)l : (off_t)sdbh.sdb_total - l + 1);
                            else 
                                osarray[counter] = curptr->data_os;
                            counter++;
                            finished++;
                        }
                    }
                    else
                    {
#endif /* USE_REGEX */
                        if (!caseflag)
                        {
                            for (fend = fstart; *fend; fend++)
                                *fend = tolower(*fend);
                        }
                        fend = strstr(fstart, match);
                        if (fend)
                        {
                            if (indexflag != 1)
                                osarray[counter] = (last_flag ? (off_t)l : (off_t)sdbh.sdb_total - l + 1);
                            else
                                osarray[counter] = curptr->data_os;
                            counter++;
                            finished++;
                        }
#ifdef USE_REGEX
                    }
#endif
                }
                else
                    finished++;		/* no more matching fields */
            }
            finished = 0;
            if (curptr)
                sdb_recfree(curptr);
        }
        if (curptr)
            free(curptr);
        curptr = NULL;
    }
#ifdef USE_REGEX
    if (rxflag)
    {
        regfree(compreg);
        free(compreg);
    }
#endif /* USE_REGEX */
    return counter;
}

/*
 * sdb_version
 *	Called by the usage() function of most tools, this function simply
 * prints (to stderr) the version of the toolkit, preceded by the name of
 * the tool (which is passed in).
 */
void
sdb_version(char *s, char *local_version)
{
    if (!s)
        return;
    if (!local_version)
        fprintf(stderr, "%s %d.%d.%d\n", s, sdb_VMAJOR, sdb_VMINOR, sdb_VPATCH);
    else
        fprintf(stderr, "%s %s / strdb %d.%d.%d\n", s, local_version, sdb_VMAJOR, sdb_VMINOR, sdb_VPATCH);
}
