/*
 * fortune.c
 *
 * This implementation of the unix-standard "fortune" program is based on
 * an underlying data file structure which differs (in some respects fairly
 * significantly) from the original and modified "strfile" formats.  For
 * details of the "strdb" format, see the core utilities in this package.
 *
 * For usage, fortune -h.  NOTE: all BSD fortune command-line switches are
 * either implemented unchanged or are faked.  Default behavior is not
 * always precisely what one expects of BSD fortune, but every attempt has
 * been made to insure that the behavior is at least to some degree
 * compatible.
 *
 * Version numbers are dual: major.minor.patch for fortune itself, and
 * major.minor.patch for the underlying strdb core.  In each case, if the
 * major version is >= 1, then the following rules apply: like the core
 * utilities, fortune will only work with data files tagged with appropriate
 * major version numbers; major version changes only with a change in the
 * data file format.  Minor version numbers indicate a change in functionality,
 * but not in persistent data structures.  Patch numbers indicate quite
 * minor changes, generally bug fixes or the addition of ports.
 * If the major version number is less than 1, we're still in beta and the
 * minor version number changes whenever data structures change, with no
 * guarantees of compatibility with other versions in any direction.
 *
 * 0.1.0/0.4.2
 *     This is the initial iteration.  It is expected to be incomplete,
 * largely used simply to outline what functionality will later be here.
 *
 * Copyright (c) 1997 Amelia A. Lewis
 * Distributed under GPL.  See COPYING.
 *
 * Note on parental responsibility and other puritan requirements:
 * This implementation of fortune doesn't give you a damned bit of help in
 * distinguishing between so-called "offensive" fortunes and those that
 * aren't.  It has capabilities built in that will permit the distinction
 * to be made, if someone wants to grapple with organizing the database
 * in that fashion.  This is not because I, the developer, don't believe
 * that there is no such thing as an offensive fortune.  It's because I
 * am offended by things that others find amusing, or merely stupid, and
 * am amused or find merely stupid some things that others are offended by.
 * MAKE UP YOUR OWN DAMN MIND.  I'm the developer, not the censor, not
 * a guardian for children brought up so incapable of independent thought
 * that they *can* be influenced by idiotic computer quote programs, not
 * a bloody miracle worker, and not b'gods likely to start taking stuff
 * out of the data files because someone's tender little feelings might
 * be *wounded*.  Deal with it.  Check the databases yourself.  It's
 * *your* damned responsibility, if you insist on taking it--I *refuse*
 * to accept it, because you're taking the trouble to compile and install
 * this piffle.
 * Flames to /dev/null.  Thank you.
 *
 * $Id: fortune.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include "../lib/strdb.h"
#include "fortune.h"

/* file names */
char filt_fn[MAXPATHLEN],	/* filter file */
     scon_fn[MAXPATHLEN],	/* system-wide configuration file */
     ucon_fn[MAXPATHLEN],	/* per-user configuration file */
     *progname;

FILE *filtf,
     *sconf,
     *uconf;

struct fort_conf
{
    char *filter;		/* defined filter */
    char *filt_params;		/* params to pass to filter */
    char *keyword;		/* keyword */
    int nfiles;			/* number of files in list */
    char **file_list;		/* list of index files */
} configs;

const char fort_VMAJOR = 0;
const char fort_VMINOR = 1;
const char fort_VPATCH = 0;

/* command line boolean switches */
int fort_babble = 0,	/* be verbose */
    fort_debug = 0,	/* print debug info */
    fort_equal = 0,	/* equalize probabilities */
    fort_long = 0,	/* use "long" fortunes */
    fort_match = 0,	/* output all matching fortunes */
    fort_nocase = 0,	/* case-insensitive matching */
    fort_report = 0,	/* just report file list */
    fort_short = 0,	/* use "short" fortunes */
    fort_wait = 0;	/* delay after display, before exit */

int fort_break = 160;	/* long/short breakpoint */


/*
 * report_defaults
 * This function is used by usage to report on compile-time defaults, such
 * as the compiled-in default name of the system-wide configuration file,
 * the per-user configuration files, and so on.
 */
void
report_defaults(void)
{
    /* default system-wide configuration file */
    /* default per-user configuration file */
    /* use of regexes */
    exit(1);
}

void
usage(char *ca, char *lv)
{
    sdb_version(ca, lv);
/* adDefF:hik:lm:n:op:stvVw */
    exit(1);
}

void
getargs(int ac, char *av[])
{
    int c, is_percent = 0;
    char **temp_flist;
    char locv[16];

    sprintf(locv, "%d.%d.%d", fort_VMAJOR, fort_VMINOR, fort_VPATCH);    
/* See, we're trying to see if we can come up with 52 command-line parameters .... */
    while ((c = getopt(ac, av, "adDefF:hik:lm:n:op:stvVw")) != EOF)
        switch (c)
        {
/* BSD */   case 'a':		/* set "ALL" keyword (default) */
                if (configs.keyword == NULL)
                {
                    configs.keyword = (char *)calloc(4, sizeof(char));
                    strcpy(configs.keyword, "ALL");
                }
                break;
            case 'd':		/* report defaults and exit */
                report_defaults();
                break;
/* BSD */   case 'D':		/* print debug info */
                fort_debug++;
                break;
/* BSD */   case 'e':		/* equalize probabilities */
                fort_equal++;
                break;
/* BSD */   case 'f':		/* report file list and exit */
                fort_report++;
                break;
            case 'F':		/* use filter named */
                if (*optarg)
                {
                    if (configs.filter)
                    {
                        free(configs.filter);
                        configs.filter = NULL;
                    }
                    configs.filter = strdup(optarg);
                }
                break;
/* BSD */   case 'i':		/* case-insensitive matching */
                fort_nocase++;
                break;
            case 'k':
                if (*optarg)
                {
                    if (configs.keyword != NULL)
                    {
                        free(configs.keyword);
                        configs.keyword = NULL;
                    }
                    configs.keyword = strdup(optarg);
                }
                break;
/* BSD */   case 'l':
                fort_long++;
                break;
/* BSD */   case 'm':		/* match regex/string */
                fort_match++;
                break;
            case 'n':		/* set short/long break point */
                if (*optarg)
                    fort_break = strtoul(optarg, NULL, 10);
                if (!fort_break)
                    fort_break = 160;
                break;
/* BSD */   case 'o':		/* set "OFF" keyword */
                if (configs.keyword == NULL)
                {
                    configs.keyword = (char *)calloc(4, sizeof(char));
                    strcpy(configs.keyword, "OFF");
                }
                break;
            case 'p':		/* pass parameter string to filter */
                if (*optarg)
                {
                    if (configs.filt_params)
                    {
                        free(configs.filt_params);
                        configs.filt_params = NULL;
                    }
                    configs.filt_params = strdup(optarg);
                }
                break;
/* BSD */   case 's':		/* short fortunes only */
                fort_short++;
                break;
            case 't':
                if (configs.keyword == NULL)
                {
                    configs.keyword = (char *)calloc(5, sizeof(char));
                    strcpy(configs.keyword, "TAME");
                }
                break;
            case 'v':		/* be verbose (??) */
                fort_babble++;
                break;
            case 'V':		/* report version and exit */
                sdb_version(progname, locv);
                exit(0);
                break;
/* BSD */   case 'w':		/* delay exit after display */
                fort_wait++;
                break;
            case 'h':		/* help! */
            case '?':
            case ':':
            default:
                sprintf(locv, "%d.%d.%d", fort_VMAJOR, fort_VMINOR, fort_VPATCH);
                usage(av[0], locv);
                break;
        }
    
    if (fort_short && fort_long)
        fort_long = 0;

    av += optind;
    while (*av)
    {
        /* check each string for whether it has a % sign -- that's a percent
           for the following file, which is a horrible thing */
        if (is_percent)
        {
            is_percent = 0;
        }
        else
        {
            if (strcasecmp(*av, "ALL") == 0)
            {
                if (configs.keyword == NULL)
                {
                    configs.keyword = (char *)calloc(4, sizeof(char));
                    strcpy(configs.keyword, "ALL");
                }
            }
            else
            {
                temp_flist = (char **)calloc(configs.nfiles + 1, sizeof(char *));
                for (c = 0; c < configs.nfiles; c++)
                {
                    temp_flist[c] = (char *)calloc(strlen(configs.file_list[c]), sizeof(char));
                    strcpy(temp_flist[c], configs.file_list[c]);
                    free(configs.file_list[c]);
                }
                strcpy(temp_flist[configs.nfiles], *av);
                configs.nfiles++;
                free(configs.file_list);
                configs.file_list = temp_flist;
                temp_flist = NULL;
            }
        }
        ++av;
    }
    return;
}

/*
 * load_config
 * Finds the system-wide configuration file; if found, loads it.  If not
 * found, loads defaults.  Then seeks per-user configuration file; if found,
 * loads it.
 */
int
load_configs(void)
{
    int keyword_plus = 0;	/* set to true if user keyword spec is addition,
                                   rather than default override */
    
    /* load the system-wide file */
    /* load the per-user file */

    /* set up the filter and parameters to it */
    if (configs.filter != NULL)
    {
        /* a filter was specified on the command line */
    }
    else
    {
        /* use the default per-user filter, if found */
        /* use the default system filter, if found */
    }
    
    if (configs.keyword != NULL)
    {
        /* set the file list from the keyword */
        /* this will *override* command line file specifications */
        /* search for it in the user config file.  May be plus. */
        /* if it wasn't there, or was plus, then look in the system file */
        /* if it wasn't there, either, use the default */
    }
    else
    {
        if (configs.nfiles)
        {
            /* files were specified on the command line.  Use them */
        }
        else
        {
            /* use the "ALL" keyword list */
            /* first check the user config file.  It might use a "plus" notation */
            /* check the system file if not found in user, or if plus */
            /* if still nothing, use the default.  This should not happen,
               since this *is* already the default action.  Misconfiguration. */
        }
    }
    return(1);
}

/*
 * collect_files
 * Probably we need to pass a char *[].  That way, we can pass in multiple
 * files, directories, or special words (like "ALL").
 *
 * This should set up some sort of global structure that has the filenames
 * and total records in each, plus a total records for all files.  It should
 * collect files in one of several ways:
 * names passed on the command line
 * directories passed on the command line, using magic to find all indices
 * symbolic names passed on the command line that refer to file collections
 *    in a configuration file (or the default symbolic names "ALL" "OFFENSIVE"
 *    and "TAME").
 * the default collection, defined in configuration files or compiled in.
 *
 * What's the return value?  Number of files?
 */
int
collect_files(void)
{
    char **temp_flist, **accum_flist, **dir_flist;
    char temp_fname[MAXPATHLEN];
    int i, ii, iii;
    int new_nfiles;
    
    /* for each element of configs.file_list, search for a matching index file */
    for (i = 0; i < configs.nfiles; i++)
    {
    /* if the "file" is a directory, collect all the index files in it; append */
    /* if no index is found, remove the entry */
    /* replace the short name with the complete pathname */
    }
    return(0);
}

void
report_files(void)
{
    int i;
    
    if (configs.nfiles)
    {
        for (i = 0; i < configs.nfiles; i++)
        {
            fprintf(stderr, "%s\n", configs.file_list[i]);
        }
    }
    return;
}

sdb_recptr *
do_it_already(unsigned long do_this)
{
    sdb_recptr *our_rec = NULL;
    
    return(our_rec);
}

unsigned long
pick_a_number(void)
{
    return(0L);
}

int
out_to_filter(sdb_recptr *orec)
{
    char *commandstring;
    int commandlen;
    
    if (!orec)
        return 0;
    if (!orec->data)
        return 0;
    if (!configs.filter)
    {
        /* try to find default output format */
        /* on failure, output raw? */
        if (!configs.filter)
        {
            printf("%s", orec->data);
        }
    }
    
    commandlen = strlen(configs.filter) + 1;
    if (configs.filt_params)
        commandlen += strlen(configs.filt_params) + 2;
    commandstring = (char *)calloc(commandlen, sizeof(char));
    strcpy(commandstring, configs.filter);
    strcat(commandstring, " ");
    strcat(commandstring, configs.filt_params);
    
    return(orec->data_size);
}

int
main(int argc, char *argv[])
{
    sdb_recptr *output_record = NULL;

    if (!load_configs())
    {
        fprintf(stderr, "%s : configuration error (useful message, eh?)\n", argv[0]);
        return 1;
    }
    getargs(argc, argv);
    if (collect_files())
    {
        /* initialize randomness */
        if (fort_report)
        {
            report_files();
            return(0);
        }
        else if (fort_match)
        {
        }
        else
        {
            output_record = do_it_already(pick_a_number());
            if (output_record)
                out_to_filter(output_record);
        }
    }
    if (fort_wait && output_record)
        sleep(output_record->data_size / 100 + 2);
    return(0);
}
