/*
 * sigstr.c
 *
 * This application is part of the strings db package.  It is a simple
 * utility which searches a particular strdb file and randomly selects
 * one item from it.  It's to be used as a random .signature file generator.
 *
 * A more ambitious application, sigd, is based loosely on the concepts
 * here, but has greater flexibility, robustness, configurability, and
 * general utility.
 *
 * version history
 * 0.1, April 10, 1997, for strdb 0.4.1.
 * 0.2, June 17, 1997/strdb 0.5.0
 *
 * Copyright (c) 1997 Amelia A. Lewis.  All rights reserved.
 * Distributed under GPL.  See ../COPYING.
 *
 * $Id: sigstr.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include "../lib/strdb.h"
#include "../lib/errdb.h"

char index_fn[MAXPATHLEN] = "",
     *progname;
FILE *indexf,
     *dataf;

int babble_flag = 0;

void
usage(char *calledas)
{
    /* it should be obvious, but this call needs to be changed, okay? */
    /* construct the version, instead. */
    sdb_version(calledas, "0.2");
    fprintf(stderr, "Usage:\n\t%s [-h] indexfile\
\n\
\n	-h		Print this usage message.\
\n	-v		Be verbose.\
\n	-V		Print version information and exit.\
\nindexfile:		Use indexfile to retrieve signatures.\n", calledas);
    exit(sdb_EUSAGE);
}

void
getargs(int ac, char *av[])
{
    int i;
    
    if (strrchr(av[0], '/'))
        progname = strdup(strrchr(av[0], '/'));
    else
        progname = strdup(av[0]);
    while ((i = getopt(ac, av, "hvV")) != EOF)
        switch (i)
        {
            case 'v':
                babble_flag++;
                break;
            case 'V':
                sdb_version(progname, "0.2");
                exit(sdb_EUSAGE);
                break;
            case 'h':
            case ':':
            case '?':
            default:
                usage(progname);
                break;
        }
    
    av += optind;
    if (*av)
        strcpy(index_fn, *av);
}

void
getfiles(sdb_header *sdbhp)
{
    int filerr;
    
    if (!index_fn[0])
    {
        usage(progname);
        exit(sdb_ENOENT);
    }
    indexf = fopen(index_fn, "r");
    if (!indexf)
    {
        perror(index_fn);
        exit(sdb_ENOENT);
    }

    filerr = sdb_unicorn(indexf, 0);
    if (filerr)
    {
        fprintf(stderr, "%s: %s", progname, sdb_err_string[filerr]);
        exit(filerr);
    }
    rewind(indexf);
    fread((sdb_header *)sdbhp, sizeof(sdb_header), 1, indexf);
    dataf = fopen(sdbhp->sdb_filename, "r");
    if (!dataf)
    {
        perror(sdbhp->sdb_filename);
        exit(sdb_ENOENT);
    }
}

int
main(int argc, char *argv[])
{
    unsigned long which;
    sdb_recptr *srp = NULL;
    sdb_header sdbh;

    srandom(time(NULL) + getpid());
    getargs(argc, argv);
    getfiles(&sdbh);

    if (babble_flag)
        fprintf(stderr, "%lu records\n", sdbh.sdb_total);
    /* get a random value in the range */
    which = 1 + (unsigned long) ((sdbh.sdb_total * 1.0) * random() / (RAND_MAX + 1.0));
    if (babble_flag)
        fprintf(stderr, "Selected record %lu\n", which);

    srp = sdb_pointer(indexf, which);
    sdb_recalloc(dataf, srp);
    fclose(indexf);
    fclose(dataf);
    
    srp->data[strlen(srp->data) - 3] = '\0';
    fputs(srp->data, stderr);
    sdb_recfree(srp);
    free(srp);

    return 0;
}
