/*
 * getkey.c
 *
 * Part of the strings db toolkit.
 *
 * getkey returns a list of the index numbers or offsets of records associated
 * with a particular key for a particular key file.  It will not return the
 * actual text of the record.
 *
 * For usage, getkey -h
 *
 * April 9, 1997: version 0.4.0, first pass.
 * 0.4.3, June 17, 1997.  Added multiple-key searches.
 *
 * Copyright (c) 1996 Amelia A. Lewis
 * Distributed under GPL.  See COPYING.
 *
 * $Id: getkey.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include "../lib/strdb.h"
#include "../lib/keydb.h"
#include "../lib/errdb.h"

char key_fn[MAXPATHLEN],
     index_fn[MAXPATHLEN],
     *progname;

FILE *keyf,
     *indexf;

int babble_flag = 0;
int gibberish_flag = 0;
int offsets_flag = 0;

unsigned char num_keys = 0;
char **search_key;

void 
usage(char *calledas)
{
    sdb_version(calledas, NULL);
    fprintf(stderr, "Usage:\n\t%s [-FhovV] file key [key ...]\n\n", calledas);
    fprintf(stderr, "\t-F\tForce use of outdated key file.\
\n	-h	Print this usage message.\
\n	-o	Report offsets instead of index numbers.\
\n	-v	Be verbose (be human-readable).\
\n	-V	Report version and exit.\
\nfile:		Name of the key file.\
\nkey:            The key(s) for which to search.\n");
    exit(1);
}

void 
getargs(int ac, char *av[])
{
    int i;

    search_key[0] = NULL;
    if (strrchr(av[0], '/'))
        progname = strdup(strrchr(av[0], '/'));
    else
        progname = strdup(av[0]);
    while ((i = getopt(ac, av, "FhovV")) != EOF)
        switch (i)
        {
            case 'F':
                gibberish_flag++;
                break;
            case 'o':
                offsets_flag++;
                break;
            case 'v':
                babble_flag++;
                break;
            case 'V':
                sdb_version(progname, NULL);
                exit(0);
                break;
            case 'h':
            case ':':
            case '?':
            default:
                usage(progname);
                break;
        }
    
    av += optind;
    if (*av)
    {
       strcpy(key_fn, *av);
    }
    else 
        usage(progname);
    if (*(av + 1))
    {
        while (*++av)
        {
            search_key[num_keys] = strdup(*av);
            num_keys++;
        }
    }
    else 
        usage(progname);
}

void 
getfiles(void)
{
    int filerr;
    sdb_key_head sdbkh;

    keyf = fopen(key_fn, "r");
    if (!keyf)
    {
        perror(key_fn);
        exit(sdb_ENOENT);
    }
    if (((filerr = sdb_unicorn(keyf, 1)) == sdb_EKEYDATE) && offsets_flag)
    {
        rewind(keyf);
        fread((sdb_key_head *)&sdbkh, sizeof(sdb_key_head), 1, keyf);
        strcpy(index_fn, sdbkh.sdb_indexfile);
        indexf = sdb_findindex(index_fn, &filerr, gibberish_flag, 0);
    }
    
    if (filerr)
    {
        fprintf(stderr, "%s", sdb_err_string[filerr]);
        exit(filerr);
    }
    if (offsets_flag)
        rewind(indexf);
    rewind(keyf);
}

sdb_key *
findkey(unsigned char which)
{
    sdb_key *fkey = NULL;
    
    fkey = (sdb_key *)malloc(sizeof(sdb_key));
    if (!fkey)
    {
        fprintf(stderr, "%s", sdb_err_string[sdb_EMEMFAULT]);
        exit(sdb_EMEMFAULT);
    }
    fkey->key = NULL;
    fkey->indices = NULL;
    fkey->nindices = 0;
    
    fseek(keyf, sizeof(sdb_key_head), SEEK_SET);

    while (!fkey->key)		/* infinite loop */
    {
        if (!sdb_key_read(fkey, keyf))	/* eof */
        {
            if (babble_flag)
                fprintf(stderr, "Key %s not found\n", search_key[which]);
            free(fkey);
            fkey = NULL;
            break;
        }
        else
        {
            if (sdb_key_match(fkey, search_key[which]))
                break;
            else
                sdb_key_free(fkey);	/* and iterate */
        }
    }
    return(fkey);
}

int
main(int argc, char *argv[])
{
    sdb_key *skey = NULL;
    sdb_recptr *rp = NULL;
    int i;
    unsigned char wk;
    
    getargs(argc, argv);
    getfiles();
    for (wk = 0; wk < num_keys; wk++)
    {
        skey = findkey(wk);
        if (skey)
        {
            if (babble_flag)
                printf("%s: %u matches\n", search_key[wk], skey->nindices);

            for (i = 0 ; i < skey->nindices ; i++)
            {
                if (offsets_flag)
                {
                    rp = sdb_pointer(indexf, skey->indices[i]);
                    if (rp)
                    {
                        if (babble_flag)
                            printf("offset %lu, size %u\n", rp->data_os, rp->data_size);
                        else
                            printf("%lu ", rp->data_os);
                    }
                    free(rp);
                    rp = NULL;
                }
                else
                {
                    if (babble_flag)
                        printf("index %u\n", skey->indices[i]);
                    else
                        printf("%u ", skey->indices[i]);
                }
            }
            if (!babble_flag)
                printf("\n");
        }
    }
    
    fclose(keyf);
    if (offsets_flag)
        fclose(indexf);
    if (skey)
        return 0;
    else
        return 2;
}
