/*
 * getstr.c
 *
 * Part of the strings db toolkit.
 *
 * getstr displays a record, indicated by index number, on stdout.
 *
 * For usage, getstr -h
 *
 * version 0.2.0, fragile but working if coddled, September 22, 1996
 * 0.4.0, April 10, 1997.  Major revision to cope with new structures.
 * 0.4.3, June 17, 1997.  Now returns multiple records when requested.
 *
 * Copyright (c) 1996 Amelia A. Lewis
 * Distributed under GPL.  See COPYING.
 *
 * $Id: getstr.c 13 2004-03-27 20:21:34Z amyzing $
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include "../lib/strdb.h"
#include "../lib/errdb.h"

char index_fn[MAXPATHLEN],
     *progname;
FILE *indexf,
     *dataf;

unsigned long recnum[255];	/* have trouble feeding the damn thing more than that on the command line */

int babble_flag = 0;
int gibberish_flag = 0;
int text_flag = 0;
int offsets_flag = 0;
int num_returns = 0;

void 
usage(char *calledas)
{
    sdb_version(calledas, NULL);
    fprintf(stderr, "Usage:\n\t%s [-fhv] file rnum [rnum ...]\n\n", calledas);
    fprintf(stderr, "\t-F\tForce use of outdated index file.\
\n	-h	Print this usage message.\
\n	-o	Input is an offset, not an index number.\
\n	-t	Display text (instead of pointer values).\
\n	-v	Be verbose, report trivial statistics.\
\n	-V	Report version of program and exit.\
\nfile:		Name of the index file.\
\nrnum:           Index of the pointer(s) to use.\n");
    exit(sdb_EUSAGE);
}

void
getargs(int ac, char *av[])
{
    int i;
    
    if (strrchr(av[0], '/'))
        progname = strdup(strrchr(av[0], '/'));
    else
        progname = strdup(av[0]);
    while ((i = getopt(ac, av, "FhotvV")) != EOF)
        switch (i)
        {
            case 'F':
                gibberish_flag++;
                break;
            case 'o':
                offsets_flag++;
                break;
            case 't':
                text_flag++;
                break;
            case 'v':
                babble_flag++;
                break;
            case 'V':
                sdb_version(progname, NULL);
                exit(0);
                break;
            case 'h':
            case ':':
            case '?':
            default:
                usage(progname);
                break;
        }
    
    av += optind;
    if (*av)
    {
       strcpy(index_fn, *av);
    }
    else 
        usage(progname);
    if (*(av + 1))
    {
        while (*++av)
        {
            recnum[num_returns] = strtoul(*av, NULL, 10);
            num_returns++;
        }
    }
    else 
        usage(progname);
}

void
getfiles(sdb_header *sdbhp)
{
    int filerr = 0;
    int i;

    indexf = sdb_findindex(index_fn, &filerr, gibberish_flag, 0);
    if (!indexf)
    {
        fprintf(stderr, "%s: %s", index_fn, sdb_err_string[filerr]);
        exit(filerr);
    }
    rewind(indexf);
    fread((sdb_header *)sdbhp, sizeof(sdb_header), 1, indexf);

    if (!offsets_flag)
        for (i = 0; i < num_returns; i++)
            if (recnum[i] > sdbhp->sdb_total)
            {
                fprintf(stderr, "%lu: %s (%lu max)\n", recnum[i], sdb_err_string[sdb_EIDXRANGE], sdbhp->sdb_total);
                exit(sdb_EIDXRANGE);
            }

    dataf = fopen(sdbhp->sdb_filename, "r");
    if (!(dataf))
    {
        perror(sdbhp->sdb_filename);
        exit(sdb_ENOENT);
    }
}

int
main(int argc, char *argv[])
{
    sdb_header sdbh;
    sdb_recptr *srp = NULL;
    char linebuf[MAXLINELEN];
    int i;
    
    getargs(argc, argv);
    getfiles(&sdbh);
    if (babble_flag)
        fprintf(stderr, "Retrieving %d records.\n", num_returns);
    else
        if (text_flag)
            fprintf(stderr, "%d records\n%c%c\n", num_returns, sdbh.sdb_sigch, sdbh.sdb_sigch);
        else
            printf("%d: ", num_returns);
    for (i = 0; i < num_returns; i++)
    {
        if (offsets_flag)
        {
            if (babble_flag)
                fprintf(stderr, "Retrieving record at offset %lu\n", recnum[i]);
            /* build srp yourself */
            srp = (sdb_recptr *)malloc(sizeof(sdb_recptr));
            srp->data_os = recnum[i];
            srp->data = NULL;
            srp->next = NULL;
            fseek(dataf, recnum[i], SEEK_SET);
            do
                fgets(linebuf, MAXLINELEN, dataf);
            while(sdb_delim(linebuf, sdbh.sdb_sigch) != 1);
            srp->data_size = ftell(dataf) - recnum[i];
        }
        else
        {
            if (babble_flag)
                fprintf(stderr, "Retrieving record %lu\n", recnum[i]);
            srp = sdb_pointer(indexf, recnum[i]);
        }
        sdb_recalloc(dataf, srp);
        if (!srp)
        {
            fprintf(stderr, "%s", sdb_err_string[sdb_EMEMFAULT]);
            exit(sdb_EMEMFAULT);
        }
        if (text_flag)
        {
            fputs(srp->data, stdout);
        }
        else
            printf("%lu %u\n", srp->data_os, srp->data_size);

        sdb_recfree(srp);
        free(srp);
    }
    fclose(dataf);
    fclose(indexf);
    return 0;
}
